#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function, unicode_literals

import os
import sys
import gzip
import base64
import struct
import hashlib
import sqlite3


"""
this reproduces incorrect sqlite3 behavior under pypy3:
 * exceptions on valid input, then succeed on retry
 * truncated SELECT results

note:
  this version was modified to use a single cursor for all operations,
  now the bug is no longer observed on any pypy version

latest version of this file:
  https://ocv.me/dev/?gist/pypy3-sqlite-bug-oneconn.py
"""


FN_DB = "up2k.db"
FN_TXT = "up2k.txt.gz"


def randsrc():
    hasher = hashlib.sha1()
    buf = b"balus"
    while True:
        hasher.update(buf)
        buf = hasher.digest()
        yield buf


def make_test_data(ntotal):
    print("creating {} lines of test data in {}".format(ntotal, FN_TXT))
    rgen = randsrc()
    with gzip.open(FN_TXT, "wb") as f:
        for n in range(ntotal):
            if n % 5000 == 0:
                print(n)

            rand_ch = base64.b32encode(next(rgen))
            randtxt = rand_ch.decode("ascii")
            rd = randtxt + randtxt
            fn = randtxt[20:]
            mt, sz = struct.unpack("<LL", rand_ch[:8])

            ln = "{} {} {}/{}\n".format(mt, sz, rd, fn)
            f.write(ln.encode("utf-8"))


class Up2k(object):
    def __init__(self):
        pass

    def build_indexes(self):
        if not os.path.exists(FN_TXT):
            make_test_data(99890)

        n_added = 0
        n_total = 0
        reproduced = False
        _db = self._open_db(FN_DB)
        db = _db.cursor()
        with gzip.open(FN_TXT, "rb") as f:
            for ln in f:
                n_total += 1
                if n_total % 5000 == 0:
                    print(n_total)

                ts, sz, rp = ln.split(b" ", 2)
                ts = int(ts)
                sz = int(sz)
                rp = rp[:-1].decode("utf-8", "replace")

                rd, fn = rp.rsplit("/", 1) if "/" in rp else ["", rp]
                sql = "select * from up where rd = ? and fn = ?"
                v = (rd, fn)
                try:
                    c = db.execute(sql, v)
                except Exception as ex:
                    print("select: " + repr(ex) + "\n  " + repr(v))
                    c = db.execute(sql, v)
                    reproduced = True

                in_db = list(c.fetchall())
                if in_db:
                    if len(in_db) > 1:
                        m = "WARN: multiple entries: [{}] |{}|\n{}"
                        rep_db = "\n".join([repr(x) for x in in_db])
                        print(m.format(rp, len(in_db), rep_db))
                    continue

                n_added += 1
                wark = hashlib.sha512((rd + "/" + fn).encode("utf-8")).digest()
                wark = base64.urlsafe_b64encode(wark)[:43]

                sql = "insert into up values (?,?,?,?,?)"
                v = (wark, ts, sz, rd, fn)
                try:
                    db.execute(sql, v)
                except Exception as ex:
                    print("insert: " + repr(ex) + "\n  " + repr(v))
                    db.execute(sql, v)
                    reproduced = True
        
        _db.commit()
        print("visited {}, added {} entries in total, {}".format(
            n_total, n_added, "REPRODUCED" if reproduced else "passed"))

        if reproduced:
            sys.exit(1)

    def _open_db(self, db_path):
        conn = sqlite3.connect(db_path)
        try:
            nfiles = next(conn.execute("select count(w) from up"))[0]
            print("found DB at {} ({} entries)".format(db_path, nfiles))
            return conn
        except:
            pass

        for cmd in [
            r"create table up (w text, mt int, sz int, rd text, fn text)",
            r"create index up_rd on up(rd)",
            r"create index up_fn on up(fn)",
        ]:
            conn.execute(cmd)

        print("created DB at {}".format(db_path))
        return conn


if __name__ == "__main__":
    if len(sys.argv) > 1:
        make_test_data(int(sys.argv[1]))
    else:
        Up2k().build_indexes()


"""
=======================================================================
sample output:  initial run  (create testdata, create database)
=======================================================================

$ pypy3 ./pypy3-sqlite-bug.py
creating 99890 lines of test data in up2k.txt.gz
0
5000
10000
15000
[...]
85000
90000
95000
created DB at up2k.db
5000
10000
15000
20000
select: InterfaceError('Error binding parameter 0 - probably unsupported type.',)
  ('LYDY46AZJ4BAIQER6UYFUJIU3WQEP5XTLYDY46AZJ4BAIQER6UYFUJIU3WQEP5XT', 'UJIU3WQEP5XT')
25000
30000
35000
[...]
85000
90000
95000
visited 99890, added 99890 new files in total, REPRODUCED
"""


"""
=======================================================================
sample output:  second run  (compare testdata to database values)
=======================================================================

$ pypy3 ./pypy3-sqlite-bug.py
found DB at up2k.db |99890|
5000
10000
15000
20000
25000
30000
WARN: multiple entries: [QUMIVCPVUG5X67EUV735X3IPINRU2Z4TQUMIVCPVUG5X67EUV735X3IPINRU2Z4T/X3IPINRU2Z4T] |2|
(b'VW8xQFnGREKqUuCNcb4_piwIx9gDZm64Gq2x2cshbkC', 1229804881, None, None, None)
(b'VW8xQFnGREKqUuCNcb4_piwIx9gDZm64Gq2x2cshbkC', 1229804881, 1448100694, 'QUMIVCPVUG5X67EUV735X3IPINRU2Z4TQUMIVCPVUG5X67EUV735X3IPINRU2Z4T', 'X3IPINRU2Z4T')
select: InterfaceError('Error binding parameter 0 - probably unsupported type.',)
  ('2PHUVYORDIQ7UUDM33HWJGKWIX2WC7X22PHUVYORDIQ7UUDM33HWJGKWIX2WC7X2', 'JGKWIX2WC7X2')
35000
40000
45000
[...]
85000
90000
95000
visited 99890, added 0 new files in total, REPRODUCED
"""
